""" Contains the legacy models of files for url redirect support. Going
forward, onegov.file and onegov.town.models.file is used.

 """

import base64

from onegov.core.filestorage import FilestorageFile


class LegacyFileCollection(object):

    def __init__(self, app):
        assert app.has_filestorage

        self.path_prefix = 'files/'
        self.file_storage = app.filestorage.makeopendir('files')

    def get_file_by_filename(self, filename):
        if self.file_storage.exists(filename):
            return LegacyFile(filename)


class LegacyImageCollection(object):

    def __init__(self, app):
        assert app.has_filestorage

        self.path_prefix = 'images/'
        self.file_storage = app.filestorage.makeopendir('images')

    def get_file_by_filename(self, filename):
        if self.file_storage.exists(filename):
            return LegacyImage(filename)


class LegacyFile(FilestorageFile):
    """ A filestorage file that points to an uploaded image or file. """

    def __init__(self, filename, info=None):
        self.filename = filename
        self.info = info or {}

    @property
    def date(self):
        if 'modified_time' in self.info:
            return self.info['modified_time'].date()

    @property
    def path(self):
        return 'files/' + self.filename

    @property
    def original_name(self):
        if '-' in self.filename:
            name = str(self.filename.split('-')[0])
            return base64.urlsafe_b64decode(name).strip()

    @classmethod
    def from_url(cls, url):
        return cls(url.split('/')[-1])


class LegacyImage(LegacyFile):
    """ A filestorage file that points to a full image (not a thumbnail). """

    @property
    def path(self):
        return 'images/' + self.filename



""" The settings view, defining things like the logo or color of the town. """

import morepath

from onegov.core.security import Private, Public
from onegov.town import _
from onegov.town.app import TownApp
from onegov.town.elements import Link
from onegov.town.forms import UserProfileForm
from onegov.town.layout import DefaultLayout
from onegov.town.models import Town
from onegov.user import UserCollection
from webob.exc import HTTPForbidden


@TownApp.form(
    model=Town, name='benutzerprofil', template='userprofile.pt',
    permission=Private, form=UserProfileForm)
def handle_user_profile(self, request, form):
    """ Handles the GET and POST login requests. """

    layout = DefaultLayout(self, request)

    collection = UserCollection(request.app.session())
    user = collection.by_username(request.identity.userid)

    if form.submitted(request):
        form.update_model(user)
        request.success(_("Your changes were saved"))
    else:
        form.apply_model(user)

    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("User Profile"), request.link(self))
    ]

    return {
        'layout': layout,
        'title': _("User Profile"),
        'form': form,
        'username': user.username,
        'initials': user.initials,
        'role': user.role
    }


# the view name must remain english, so that automated tools can detect it
@TownApp.html(model=Town, name='unsubscribe', template='userprofile.pt',
              permission=Public)
def handle_unsubscribe(self, request):
    """ Unsubscribes a user from all *regular* e-mails.

    To be able to use this method, an url has to be created like this::

        '{}?token={}'.format((
            request.link(town, name='unsubscribe'),
            request.new_url_safe_token(
                {'user': 'user@example.org'}, 'unsubscribe'
            )
        ))

    """

    # tokens are valid for 30 days
    max_age = 60 * 60 * 24 * 30
    salt = 'unsubscribe'
    newsletters = {
        'daily_ticket_statistics'
    }

    data = request.load_url_safe_token(
        request.params.get('token'), max_age=max_age, salt=salt
    )

    if not data:
        return HTTPForbidden()

    user = UserCollection(request.app.session()).by_username(data['user'])

    if not user:
        return HTTPForbidden()

    if not user.data:
        user.data = {}

    for newsletter in newsletters:
        user.data[newsletter] = False

    request.success(
        _("You have been successfully unsubscribed from all regular emails.")
    )
    return morepath.redirect(request.link(self))



import json
import morepath
import sedate

from collections import OrderedDict, namedtuple
from datetime import datetime, timedelta
from isodate import parse_date, ISO8601Error
from itertools import groupby
from libres.db.models import Reservation
from morepath.request import Response
from onegov.core.csv import convert_list_of_dicts_to_csv
from onegov.core.csv import convert_list_of_dicts_to_xlsx
from onegov.core.security import Public, Private
from onegov.core.utils import normalize_for_url
from onegov.form import FormSubmission
from onegov.libres import ResourceCollection
from onegov.libres.models import Resource
from onegov.ticket import Ticket
from onegov.town import TownApp, _
from onegov.town import utils
from onegov.town.elements import Link
from onegov.town.forms import (
    ResourceForm, ResourceCleanupForm, ResourceExportForm
)
from onegov.town.layout import ResourcesLayout, ResourceLayout
from onegov.town.models.resource import DaypassResource, RoomResource
from sqlalchemy.sql.expression import nullsfirst
from webob import exc


RESOURCE_TYPES = {
    'daypass': {
        'success': _("Added a new daypass"),
        'title': _("New daypass"),
        'class': DaypassResource
    },
    'room': {
        'success': _("Added a new room"),
        'title': _("New room"),
        'class': RoomResource
    }
}


def get_daypass_form(self, request):
    return get_resource_form(self, request, 'daypass')


def get_room_form(self, request):
    return get_resource_form(self, request, 'room')


def get_resource_form(self, request, type=None):
    if isinstance(self, ResourceCollection):
        assert type is not None
        model = RESOURCE_TYPES[type]['class']()
    else:
        model = self

    return model.with_content_extensions(ResourceForm, request)


@TownApp.html(model=ResourceCollection, template='resources.pt',
              permission=Public)
def view_resources(self, request):
    resources = self.query().order_by(nullsfirst(Resource.group)).all()
    resources = request.exclude_invisible(resources)
    grouped = OrderedDict()

    for group, items in groupby(resources, lambda r: r.group or _("General")):
        grouped[group] = sorted(tuple(items), key=lambda r: r.title)

    if len(grouped) == 1:
        grouped = {None: tuple(grouped.values())[0]}

    return {
        'title': _("Reservations"),
        'resources': grouped,
        'layout': ResourcesLayout(self, request)
    }


@TownApp.form(model=ResourceCollection, name='neuer-raum',
              template='form.pt', permission=Private, form=get_room_form)
def handle_new_room(self, request, form):
    return handle_new_resource(self, request, form, 'room')


@TownApp.form(model=ResourceCollection, name='neue-tageskarte',
              template='form.pt', permission=Private, form=get_daypass_form)
def handle_new_daypass(self, request, form):
    return handle_new_resource(self, request, form, 'daypass')


def handle_new_resource(self, request, form, type):
    if form.submitted(request):

        resource = self.add(
            title=form.title.data, type=type, timezone='Europe/Zurich'
        )
        form.populate_obj(resource)

        request.success(RESOURCE_TYPES[type]['success'])
        return morepath.redirect(request.link(resource))

    layout = ResourcesLayout(self, request)
    layout.include_editor()
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(RESOURCE_TYPES[type]['title'], '#'))

    return {
        'layout': layout,
        'title': _(RESOURCE_TYPES[type]['title']),
        'form': form,
        'form_width': 'large'
    }


@TownApp.form(model=Resource, name='bearbeiten', template='form.pt',
              permission=Private, form=get_resource_form)
def handle_edit_resource(self, request, form):
    if form.submitted(request):
        form.populate_obj(self)

        request.success(_("Your changes were saved"))
        return morepath.redirect(request.link(self))

    elif not request.POST:
        form.process(obj=self)

    layout = ResourceLayout(self, request)
    layout.include_editor()
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(_("Edit"), '#'))

    return {
        'layout': layout,
        'title': self.title,
        'form': form,
        'form_width': 'large'
    }


@TownApp.html(model=Resource, template='resource.pt', permission=Public)
def view_resource(self, request):
    return {
        'title': self.title,
        'resource': self,
        'layout': ResourceLayout(self, request),
        'feed': request.link(self, name='slots')
    }


@TownApp.view(model=Resource, request_method='DELETE', permission=Private)
def handle_delete_resource(self, request):

    if not self.deletable:
        raise exc.HTTPMethodNotAllowed()

    collection = ResourceCollection(request.app.libres_context)
    collection.delete(self, including_reservations=False)


@TownApp.form(model=Resource, permission=Private, name='cleanup',
              form=ResourceCleanupForm, template='resource_cleanup.pt')
def handle_cleanup_allocations(self, request, form):
    """ Removes all unused allocations between the given dates. """

    if form.submitted(request):
        start, end = form.data['start'], form.data['end']
        count = self.scheduler.remove_unused_allocations(start, end)

        request.success(
            _("Successfully removed ${count} unused allocations", mapping={
                'count': count
            })
        )

        return morepath.redirect(request.link(self))

    if request.method == 'GET':
        form.start.data, form.end.data = get_date_range(self, request.params)

    layout = ResourceLayout(self, request)
    layout.breadcrumbs.append(Link(_("Clean up"), '#'))
    layout.editbar_links = None

    return {
        'layout': layout,
        'title': _("Clean up"),
        'form': form
    }


@TownApp.json(model=Resource, name='reservations', permission=Public)
def get_reservations(self, request):
    return [
        utils.ReservationInfo(reservation, request).as_dict() for reservation
        in self.bound_reservations(request).all()
    ]


def get_date(text, default):
    try:
        date = parse_date(text)
        return datetime(date.year, date.month, date.day, tzinfo=default.tzinfo)
    except (ISO8601Error, TypeError):
        return default


def get_date_range(resource, params):
    default_start, default_end = resource.calendar_date_range

    start = get_date(params.get('start'), default_start)
    end = get_date(params.get('end'), default_end)

    start = sedate.replace_timezone(
        datetime(start.year, start.month, start.day), resource.timezone)

    end = sedate.replace_timezone(
        datetime(end.year, end.month, end.day), resource.timezone)

    if end < start:
        start = end

    return sedate.align_range_to_day(start, end, resource.timezone)


@TownApp.html(model=Resource, permission=Private, name='belegung',
              template='resource_occupancy.pt')
def view_occupancy(self, request):

    # infer the default start/end date from the calendar view parameters
    start, end = get_date_range(self, request.params)

    query = self.reservations_with_tickets_query(start, end)
    query = query.with_entities(
        Reservation.start, Reservation.end, Reservation.quota,
        Ticket.subtitle, Ticket.id
    )

    def group_key(record):
        return sedate.to_timezone(record[0], self.timezone).date()

    occupancy = OrderedDict()
    grouped = groupby(query.all(), group_key)
    Entry = namedtuple('Entry', ('start', 'end', 'title', 'quota', 'url'))
    count = 0

    for date, records in grouped:
        occupancy[date] = tuple(
            Entry(
                start=sedate.to_timezone(r[0], self.timezone),
                end=sedate.to_timezone(
                    r[1] + timedelta(microseconds=1), self.timezone),
                quota=r[2],
                title=r[3],
                url=request.class_link(Ticket, {
                    'handler_code': 'RSV',
                    'id': r[4]
                })
            ) for r in records
        )
        count += len(occupancy[date])

    layout = ResourceLayout(self, request)
    layout.breadcrumbs.append(Link(_("Occupancy"), '#'))
    layout.editbar_links = None

    return {
        'layout': layout,
        'title': _("Occupancy"),
        'occupancy': occupancy,
        'resource': self,
        'start': sedate.to_timezone(start, self.timezone).date(),
        'end': sedate.to_timezone(end, self.timezone).date(),
        'count': count,
    }


@TownApp.form(model=Resource, permission=Private, name='export',
              template='resource_export.pt', form=ResourceExportForm)
def view_export(self, request, form):

    # XXX this could be turned into a redirect to a GET view, which would
    # make it easier for scripts to get this data, but since we don't have
    # a good API story anyway we don't have spend to much energy on it here
    # - instead we should do this in a comprehensive fashion
    if form.submitted(request):
        file_format = form.data['file_format']

        constant_fields, results = run_export(
            resource=self,
            request=request,
            start=form.data['start'],
            end=form.data['end'],
            nested=file_format == 'json'
        )

        def field_order(field):
            # known fields come first in a defined order (-x -> -1)
            # unknown fields are ordered from a-z (second item in tuple)
            if field in constant_fields:
                return constant_fields.index(field) - len(constant_fields), ''
            else:
                return 0, field

        if file_format == 'json':
            return Response(
                json.dumps(results),
                content_type='application/json'
            )
        elif file_format == 'csv':
            return Response(
                convert_list_of_dicts_to_csv(results, key=field_order),
                content_type='text/plain'
            )
        elif file_format == 'xlsx':
            return Response(
                convert_list_of_dicts_to_xlsx(results, key=field_order),
                content_type=(
                    'application/vnd.openxmlformats'
                    '-officedocument.spreadsheetml.sheet'
                ),
                content_disposition='inline; filename={}.xlsx'.format(
                    normalize_for_url(self.title)
                )
            )
        else:
            raise NotImplemented()

    if request.method == 'GET':
        form.start.data, form.end.data = get_date_range(self, request.params)

    layout = ResourceLayout(self, request)
    layout.breadcrumbs.append(Link(_("Occupancy"), '#'))
    layout.editbar_links = None

    return {
        'layout': layout,
        'title': _("Export"),
        'form': form
    }


def run_export(resource, request, start, end, nested):
    start = sedate.replace_timezone(
        datetime(start.year, start.month, start.day),
        resource.timezone
    )
    end = sedate.replace_timezone(
        datetime(end.year, end.month, end.day),
        resource.timezone
    )

    start, end = sedate.align_range_to_day(start, end, resource.timezone)

    query = resource.reservations_with_tickets_query(start, end)
    query = query.join(FormSubmission, Reservation.token == FormSubmission.id)
    query = query.with_entities(
        Reservation.start,
        Reservation.end,
        Reservation.quota,
        Reservation.email,
        Ticket.number,
        Ticket.subtitle,
        FormSubmission.data,
    )

    results = []

    # update me: reused outside this function
    constant_fields = ('start', 'end', 'quota', 'email', 'ticket', 'title')

    for record in query.all():
        result = OrderedDict()

        start = sedate.to_timezone(record[0], resource.timezone)
        end = sedate.to_timezone(record[1], resource.timezone)
        end += timedelta(microseconds=1)

        result['start'] = start.isoformat()
        result['end'] = end.isoformat()
        result['quota'] = record[2]
        result['email'] = record[3]
        result['ticket'] = record[4]
        result['title'] = record[5]

        if nested:
            result['form'] = record[6]
        else:
            for key, value in record[6].items():
                if isinstance(value, list):
                    result['form_' + key] = ', '.join(value)
                else:
                    result['form_' + key] = value

        results.append(result)

    return constant_fields, results



""" The newsletter view. """

import morepath

from collections import OrderedDict
from itertools import groupby
from onegov.core.security import Public, Private
from onegov.core.templates import render_template
from onegov.core.utils import linkify
from onegov.event import Occurrence, OccurrenceCollection
from onegov.newsletter import (
    Newsletter,
    NewsletterCollection,
    Recipient,
    RecipientCollection
)
from onegov.newsletter.errors import AlreadyExistsError
from onegov.town import _, TownApp
from onegov.town.forms import NewsletterForm, NewsletterSendForm, SignupForm
from onegov.town.layout import (
    DefaultMailLayout,
    NewsletterLayout,
    RecipientLayout
)
from onegov.town.models import News
from sedate import utcnow
from sqlalchemy import desc
from sqlalchemy.orm import undefer


def get_newsletter_form(model, request):

    form = NewsletterForm

    query = request.app.session().query(News)
    query = query.filter(News.parent != None)
    query = query.order_by(desc(News.created))
    query = query.options(undefer('created'))
    form = form.with_news(request, query.all())

    query = OccurrenceCollection(request.app.session()).query(outdated=False)
    subquery = query.with_entities(Occurrence.id)
    subquery = subquery.order_by(None)
    subquery = subquery.order_by(
        Occurrence.event_id,
        Occurrence.start
    )
    subquery = subquery.distinct(Occurrence.event_id).subquery()
    query.filter(Occurrence.id.in_(subquery))

    form = form.with_occurrences(request, query.all())

    return form


def get_newsletter_send_form(model, request):
    query = RecipientCollection(request.app.session()).query()
    query = query.order_by(Recipient.address)

    return NewsletterSendForm.for_newsletter(model, query.all())


def news_by_newsletter(newsletter, request):
    news_ids = newsletter.content.get('news')

    if not news_ids:
        return None

    query = request.app.session().query(News)
    query = query.order_by(desc(News.created))
    query = query.options(undefer('created'))
    query = query.options(undefer('content'))
    query = query.filter(News.id.in_(news_ids))

    return request.exclude_invisible(query.all())


def occurrences_by_newsletter(newsletter, request):
    occurrence_ids = newsletter.content.get('occurrences')

    if not occurrence_ids:
        return None

    query = request.app.session().query(Occurrence)
    query = query.order_by(Occurrence.start, Occurrence.title)
    query = query.filter(Occurrence.id.in_(occurrence_ids))

    return query.all()


@TownApp.form(model=NewsletterCollection, template='newsletter_collection.pt',
              permission=Public, form=SignupForm)
def handle_newsletters(self, request, form):

    if form.submitted(request):
        recipients = RecipientCollection(request.app.session())
        recipient = recipients.by_address(form.address.data)

        # do not show a specific error message if the user already signed up,
        # just pretend like everything worked correctly - if someone signed up
        # or not is private

        if not recipient:
            recipient = recipients.add(address=form.address.data)

            title = request.translate(
                _("Welcome to the ${town} Newsletter", mapping={
                    'town': request.app.town.name
                })
            )

            confirm_mail = render_template('mail_confirm.pt', request, {
                'layout': DefaultMailLayout(self, request),
                'newsletters': self,
                'subscription': recipient.subscription,
                'title': title
            })

            request.app.send_email(
                subject=title,
                receivers=(recipient.address, ),
                content=confirm_mail
            )

        request.success(_((
            "Success! We have sent a confirmation link to "
            "${address}, if we didn't send you one already."
        ), mapping={'address': form.address.data}))

    query = self.query()
    query = query.options(undefer(Newsletter.created))
    query = query.order_by(desc(Newsletter.created))

    # newsletters which were not sent yet are private
    if not request.is_logged_in:
        query = query.filter(Newsletter.sent != None)

    # the recipients count is only shown to logged in users
    if request.is_logged_in:
        recipients_count = RecipientCollection(self.session).query()\
            .filter(Recipient.confirmed == True)\
            .count()
    else:
        recipients_count = 0

    return {
        'form': form,
        'layout': NewsletterLayout(self, request),
        'newsletters': query.all(),
        'title': _("Newsletter"),
        'recipients_count': recipients_count
    }


@TownApp.html(model=Newsletter, template='newsletter.pt', permission=Public)
def view_newsletter(self, request):

    return {
        'layout': NewsletterLayout(self, request),
        'newsletter': self,
        'news': news_by_newsletter(self, request),
        'occurrences': occurrences_by_newsletter(self, request),
        'title': self.title,
        'lead': linkify(self.lead),
    }


@TownApp.html(model=RecipientCollection, template='recipients.pt',
              permission=Private)
def view_subscribers(self, request):

    # i18n:attributes translations do not support variables, so we need
    # to do this ourselves
    warning = request.translate(_("Do you really want to unsubscribe \"{}\"?"))

    recipients = self.query().order_by(Recipient.address).all()
    by_letter = OrderedDict()

    for key, values in groupby(recipients, key=lambda r: r.address[0].upper()):
        by_letter[key] = list(values)

    return {
        'layout': RecipientLayout(self, request),
        'title': _("Subscribers"),
        'by_letter': by_letter,
        'warning': warning,
    }


@TownApp.form(model=NewsletterCollection, name='neu', template='form.pt',
              permission=Public, form=get_newsletter_form)
def handle_new_newsletter(self, request, form):

    if form.submitted(request):
        try:
            newsletter = self.add(title=form.title.data, html='')
        except AlreadyExistsError:
            request.alert(_("A newsletter with this name already exists"))
        else:
            form.update_model(newsletter, request)

            request.success(_("Added a new newsletter"))
            return morepath.redirect(request.link(newsletter))

    return {
        'form': form,
        'layout': NewsletterLayout(self, request),
        'title': _("New Newsletter"),
        'size': 'large'
    }


@TownApp.form(model=Newsletter, template='form.pt', name='bearbeiten',
              permission=Private, form=get_newsletter_form)
def edit_newsletter(self, request, form):

    if form.submitted(request):
        form.update_model(self, request)

        request.success(_("Your changes were saved"))
        return morepath.redirect(request.link(self))

    elif request.method == 'GET':
        form.apply_model(self)

    return {
        'layout': NewsletterLayout(self, request),
        'form': form,
        'title': _("Edit Newsletter"),
        'size': 'large'
    }


@TownApp.view(model=Newsletter, request_method='DELETE', permission=Private)
def delete_page(self, request):
    request.assert_valid_csrf_token()

    NewsletterCollection(request.app.session()).delete(self)
    request.success("The newsletter was deleted")


@TownApp.form(model=Newsletter, template='send_newsletter.pt', name='senden',
              permission=Private, form=get_newsletter_send_form)
def handle_send_newsletter(self, request, form):

    if form.submitted(request):
        query = RecipientCollection(request.app.session()).query()
        query = query.filter(Recipient.id.in_(form.recipients.data))
        recipients = (r for r in query.all() if r not in self.recipients)

        sent_mails = 0

        for recipient in recipients:

            # one mail per recipient (each has a specific unsubscribe link)
            mail = render_template(
                'mail_newsletter.pt', request, {
                    'layout': DefaultMailLayout(self, request),
                    'lead': linkify(self.lead).replace('\n', '<br>'),
                    'newsletter': self,
                    'title': self.title,
                    'unsubscribe': request.link(
                        recipient.subscription,
                        'unsubscribe'
                    ),
                    'news': news_by_newsletter(self, request),
                    'occurrences': occurrences_by_newsletter(self, request),
                }
            )

            request.app.send_email(
                subject=self.title,
                receivers=(recipient.address, ),
                content=mail
            )

            self.recipients.append(recipient)
            sent_mails += 1

        if not self.sent:
            self.sent = utcnow()

        request.success(_('Sent "${title}" to ${n} recipients', mapping={
            'title': self.title,
            'n': sent_mails
        }))
        return morepath.redirect(request.link(self))

    return {
        'layout': NewsletterLayout(self, request),
        'form': form,
        'title': self.title,
        'newsletter': self,
        'previous_recipients': self.recipients
    }



from collections import OrderedDict

from onegov.foundation import BaseTheme
from onegov.core.utils import module_path


# options editable by the user
user_options = {
    'primary-color': '#006fba',
    'footer-height': '38px',
    'include-print-styles': 'false'
}


class TownTheme(BaseTheme):
    name = 'onegov.town.foundation'

    # don't touch this number, it's incremented using bumpversion, so every
    # release will automatically trigger a rebuild of the theme
    version = '1.10.3'

    @property
    def default_options(self):
        options = OrderedDict((
            # base colors
            ('gray', '#e0e3e5'),
            ('black', '#0f0f0f'),
            ('blue', '#2575ed'),
            ('red', '#de2c3b'),
            ('yellow', '#ffc800'),
            ('orange', '#ffb100'),
            ('green', '#2c9f42'),
            ('white', '#fff'),
            ('gray-dark', '#b3b6b7'),
            ('black-dark', '#000'),
            ('blue-dark', '#1a52a5'),
            ('red-dark', '#b2232f'),
            ('yellow-dark', '#cca000'),
            ('orange-dark', '#d59200'),
            ('green-dark', '#237f35'),
            ('gray-light', '#f0f1f2'),
            ('black-light', '#363738'),
            ('blue-light', '#92baf6'),
            ('red-light', '#ef969d'),
            ('yellow-light', '#ffe480'),
            ('green-light', '#96cfa1'),
            ('gray-pastel', '#f7f8f8'),
            ('black-pastel', '#dadada'),
            ('blue-pastel', '#d3e3fb'),
            ('red-pastel', '#f8d5d8'),
            ('yellow-pastel', '#fff4cc'),
            ('green-pastel', '#d4ecd8'),

            # zurb overrides
            ('alert-color', '$red'),
            ('success-color', '$green'),
            ('warning-color', '$yellow'),
            ('info-color', '$blue'),
            ('callout-panel-bg', '$yellow-light'),
            ('callout-panel-border-color', '$yellow'),
            ('top-bar-border-size', '0.3rem'),
            ('bottom-links-color', '#777'),
            ('bottom-links-size', '0.8rem'),
            ('topbar-bg-color', '$gray-pastel'),
            ('topbar-link-bg-hover', '$gray'),
            ('topbar-link-color', '#312f2e'),
            ('topbar-link-color-hover', '#312f2e'),
            ('topbar-link-color-active', '#312f2e'),
            ('topbar-link-color-active-hover', '#312f2e'),
            ('topbar-link-weight', 'bold'),
            ('topbar-menu-link-color', '#312f2e'),
            ('topbar-menu-icon-color', '#312f2e'),
            ('topbar-dropdown-bg', '$gray-pastel'),
            ('side-nav-font-weight-active', 'bold'),
            ('side-nav-padding', '0'),
            ('side-nav-link-padding', 'rem-calc(0 14)'),
            ('crumb-bg', '#fff'),
            ('crumb-border-size', '0'),
            ('header-line-height', '1.25'),
            ('sub-nav-padding', '.25rem'),
            ('table-border-style', 'none'),
            ('thumb-border-width', '0'),
            ('thumb-box-shadow', '0 0 0 2px rgba($black,.1)'),
            ('thumb-box-shadow-hover', '0 0 0 2px rgba($black,.3)'),

            # custom
            ('tile-image-1', '"../static/homepage-images/tile-1-small.jpg"'),
            ('tile-image-2', '"../static/homepage-images/tile-2-small.jpg"'),
            ('tile-image-3', '"../static/homepage-images/tile-3-small.jpg"'),
            ('tile-image-4', '"../static/homepage-images/tile-4-small.jpg"'),
            ('tile-image-5', '"../static/homepage-images/tile-5-small.jpg"'),
            ('tile-image-6', '"../static/homepage-images/tile-6-small.jpg"'),
        ))
        options.update(user_options)

        return options

    @property
    def pre_imports(self):
        return [
            'foundation-mods'
        ]

    @property
    def post_imports(self):
        return [
            'town'
        ]

    @property
    def extra_search_paths(self):
        return [module_path('onegov.town.theme', 'styles')]



from onegov.core.utils import sanitize_html
from onegov.form import Form
from onegov.form.validators import ValidFormDefinition
from onegov.town import _
from onegov.town.forms.reservation import RESERVED_FIELDS
from onegov.town.utils import annotate_html
from wtforms import RadioField, StringField, TextAreaField, validators
from wtforms.fields.html5 import DateField


class ResourceForm(Form):
    """ Defines the form for all resources. """
    title = StringField(_("Title"), [validators.InputRequired()])

    lead = TextAreaField(
        label=_("Lead"),
        description=_("Describes what this reservation resource is about"),
        render_kw={'rows': 4})

    group = StringField(
        label=_("Group"),
        description=_("Used to group the resource in the overview")
    )

    text = TextAreaField(
        label=_("Text"),
        render_kw={'class_': 'editor'},
        filters=[sanitize_html, annotate_html])

    definition = TextAreaField(
        label=_("Extra Fields Definition"),
        validators=[
            validators.Optional(),
            ValidFormDefinition(
                require_email_field=False,
                reserved_fields=RESERVED_FIELDS
            )
        ],
        render_kw={'rows': 32, 'data-editor': 'form'}
    )


class DateRangeForm(Form):
    """ A form providing a start/end date range. """

    start = DateField(
        label=_("Start"),
        validators=[validators.InputRequired()]
    )

    end = DateField(
        label=_("End"),
        validators=[validators.InputRequired()]
    )

    def validate(self):
        result = super().validate()

        if self.start.data and self.end.data:
            if self.start.data > self.end.data:
                message = _("The end date must be later than the start date")
                self.end.errors.append(message)
                result = False

        return result


class ResourceCleanupForm(DateRangeForm):
    """ Defines the form to remove multiple allocations. """


class ResourceExportForm(DateRangeForm):
    """ Defines the form to export reservations. """

    file_format = RadioField(
        label=_("Format"),
        choices=[
            ('csv', _("CSV File")),
            ('xlsx', _("Excel File")),
            ('json', _("JSON File"))
        ],
        default='csv',
        validators=[
            validators.InputRequired()
        ]
    )



from onegov.form import Form
from onegov.town import _
from wtforms.fields.html5 import EmailField
from wtforms.validators import Email, InputRequired


# include all fields used below so we can filter them out
# when we merge this form with the custom form definition
RESERVED_FIELDS = ['email']


class ReservationForm(Form):

    reserved_fields = RESERVED_FIELDS

    email = EmailField(
        label=_("E-Mail"),
        validators=[InputRequired(), Email()]
    )



""" Contains a list of manual translations. Those are usually messages that
exist in external packages which provide no translations.

"""

from onegov.town import _

# messages defined in wtforms-components
_("Not a valid color.")



import babel.dates
import sedate

from cached_property import cached_property
from datetime import timedelta
from dateutil import rrule
from onegov.core.layout import ChameleonLayout
from onegov.core.static import StaticFile
from onegov.core.crypto import RANDOM_TOKEN_LENGTH
from onegov.core.utils import linkify
from onegov.event import OccurrenceCollection
from onegov.form import FormCollection, FormSubmissionFile, render_field
from onegov.libres import ResourceCollection
from onegov.page import Page, PageCollection
from onegov.people import PersonCollection
from onegov.ticket import TicketCollection
from onegov.town import _
from onegov.town import utils
from onegov.town.elements import DeleteLink, Link, LinkGroup
from onegov.town.models import (
    GeneralFileCollection,
    ImageFile,
    ImageFileCollection,
    ImageSetCollection,
    PageMove,
    PersonMove,
    Search,
    SiteCollection
)
from onegov.town.models.extensions import PersonLinkExtension
from onegov.town.theme.town_theme import user_options
from onegov.newsletter import NewsletterCollection, RecipientCollection
from onegov.user import Auth
from purl import URL
from sqlalchemy import desc


class Layout(ChameleonLayout):
    """ Contains methods to render a page inheriting from layout.pt.

    All pages inheriting from layout.pt rely on this class being present
    as 'layout' variable::

     @TownApp.html(model=Example, template='example.pt', permission=Public)
        def view_example(self, request):
            return { 'layout': DefaultLayout(self, request) }

    It is meant to be extended for different parts of the site. For example,
    the :class:`DefaultLayout` includes the top navigation defined by
    onegov.page.

    It's possible though to have a different part of the website use a
    completely different top navigation. For that, a new Layout class
    inheriting from this one should be added.

    """

    @property
    def town(self):
        """ An alias for self.request.app.town. """
        return self.request.app.town

    @property
    def view_name(self):
        """ Returns the view name of the current view, or None if it is the
        default view.

        Note: This relies on morepath internals and is experimental in nature!

        """
        return self.request.unconsumed and self.request.unconsumed[-1] or None

    @property
    def primary_color(self):
        return self.town.theme_options.get(
            'primary-color', user_options['primary-color'])

    def today(self):
        return sedate.to_timezone(sedate.utcnow(), self.timezone).date()

    @cached_property
    def default_map_view(self):
        return self.town.default_map_view or None

    @cached_property
    def svg(self):
        return self.template_loader['svg.pt']

    @cached_property
    def font_awesome_path(self):
        static_file = StaticFile.from_application(
            self.app, 'font-awesome/css/font-awesome.min.css')

        return self.request.link(static_file)

    @cached_property
    def page_id(self):
        """ Returns the unique page id of the rendered page. Used to have
        a useful id in the body element for CSS/JS.

        """
        page_id = self.request.path_info
        page_id = page_id.lstrip('/')
        page_id = page_id.replace('/', '-')
        page_id = page_id.rstrip('-')

        return page_id or 'root'

    @cached_property
    def top_navigation(self):
        """ Returns a list of :class:`onegov.town.elements.Link` objects.
        Those links are used for the top navigation.

        If nothing is returned, no top navigation is displayed.

        """
        return None

    @cached_property
    def breadcrumbs(self):
        """ Returns a list of :class:`onegov.town.elements.Link` objects.
        Those links are used for the breadcrumbs.

        If nothing is returned, no top breadcrumbs are displayed.

        """
        return None

    @cached_property
    def sidebar_links(self):
        """ A list of links shown in the sidebar, used for navigation. """
        return None

    @cached_property
    def bottom_links(self):
        """ A list of links shown at the absolute bottom. Use this for
        links like administration, statistics, source-code.

        """
        return None

    @cached_property
    def editbar_links(self):
        """ A of :class:`onegov.town.elements.LinkGroup` classes. Each of them
        will be shown in the top editbar, with the group title being the
        dropdown title.
        """
        return None

    @cached_property
    def file_upload_url(self):
        """ Returns the url to the file upload action. """
        url = self.request.link(GeneralFileCollection(self.app), name='upload')
        return self.csrf_protected_url(url)

    @cached_property
    def file_upload_json_url(self):
        """ Adds the json url for file uploads. """
        url = self.request.link(
            GeneralFileCollection(self.app), name='upload.json'
        )
        return self.csrf_protected_url(url)

    @cached_property
    def file_list_url(self):
        """ Adds the json url for file lists. """
        return self.request.link(GeneralFileCollection(self.app), name='json')

    @cached_property
    def image_upload_url(self):
        """ Returns the url to the image upload action. """
        url = self.request.link(ImageFileCollection(self.app), name='upload')
        return self.csrf_protected_url(url)

    @cached_property
    def image_upload_json_url(self):
        """ Adds the json url for image uploads. """
        url = self.request.link(
            ImageFileCollection(self.app), name='upload.json'
        )
        return self.csrf_protected_url(url)

    @cached_property
    def image_list_url(self):
        """ Adds the json url for image lists. """
        return self.request.link(ImageFileCollection(self.app), name='json')

    @cached_property
    def sitecollection_url(self):
        """ Adds the json url for internal links lists. """
        return self.request.link(SiteCollection(self.app.session()))

    @cached_property
    def homepage_url(self):
        """ Returns the url to the main page. """
        return self.request.link(self.app.town)

    @cached_property
    def search_url(self):
        """ Returns the url to the search page. """
        return self.request.link(Search(self.request, None, None))

    @cached_property
    def suggestions_url(self):
        """ Returns the url to the suggestions json view. """
        return self.request.link(Search(self.request, None, None), 'suggest')

    @cached_property
    def open_tickets_url(self):
        return self.request.link(
            TicketCollection(self.request.app.session(), state='open'))

    @cached_property
    def pending_tickets_url(self):
        return self.request.link(
            TicketCollection(self.request.app.session(), state='pending'))

    @cached_property
    def closed_tickets_url(self):
        return self.request.link(
            TicketCollection(self.request.app.session(), state='closed'))

    @cached_property
    def login_url(self):
        """ Returns the login url for the current page. """
        return self.request.link(
            Auth.from_app(self.app, self.request.transform(self.request.path)),
            name='login'
        )

    @cached_property
    def logout_url(self):
        """ Returns the logout url for the current page. """
        return self.request.link(
            Auth.from_app(self.app, self.request.transform(self.request.path)),
            name='logout'
        )

    @cached_property
    def events_url(self):
        return self.request.link(
            OccurrenceCollection(self.request.app.session())
        )

    def thumbnail_url(self, url):
        """ Takes the given url and returns the thumbnail url for it.

        Uses some rough heuristic to determine if a url is actually served
        by onegov.file or not. May possibly fail.

        """
        if '/storage/' not in url:
            return url

        image_id = url.split('/storage/')[-1]

        # image file ids are generated from the random_token function
        if len(image_id) == RANDOM_TOKEN_LENGTH:
            return self.request.class_link(
                ImageFile, {'id': image_id}, name='thumbnail')
        else:
            return url

    def include_editor(self):
        self.request.include('redactor')
        self.request.include('editor')

    def include_code_editor(self):
        self.request.include('code_editor')

    def render_field(self, field):
        """ Alias for ``onegov.form.render_field``. """
        return render_field(field)

    def field_download_link(self, field):
        if not field.type == 'UploadField':
            return None

        if field.data.get('data', '').startswith('@'):
            return self.request.link(
                FormSubmissionFile(id=field.data['data'].lstrip('@')))

    @cached_property
    def move_person_url_template(self):
        assert isinstance(self.model, PersonLinkExtension)

        implementation = PersonMove.get_implementation(self.model)
        move = implementation.for_url_template(self.model)

        return self.csrf_protected_url(self.request.link(move))

    def get_user_color(self, username):
        return utils.get_user_color(username)

    def format_time_range(self, start, end):
        return utils.format_time_range(start, end)

    def format_timedelta(self, delta):
        return babel.dates.format_timedelta(
            delta=delta,
            locale=self.request.locale
        )

    def format_seconds(self, seconds):
        return self.format_timedelta(timedelta(seconds=seconds))


class DefaultMailLayout(Layout):
    """ A special layout for creating HTML E-Mails. """

    @cached_property
    def base(self):
        return self.template_loader['mail_layout.pt']

    @cached_property
    def macros(self):
        return self.template_loader['mail_macros.pt']

    @cached_property
    def contact_html(self):
        """ Returns the contacts html, but instead of breaking it into multiple
        lines (like on the site footer), this version puts it all on one line.

        """

        lines = (l.strip() for l in self.town.meta['contact'].splitlines())
        lines = (l for l in lines if l)

        return linkify(', '.join(lines))

    def unsubscribe_link(self, username):
        return '{}?token={}'.format(
            self.request.link(self.town, name='unsubscribe'),
            self.request.new_url_safe_token(
                data={'user': username},
                salt='unsubscribe'
            )
        )


class DefaultLayout(Layout):
    """ The defaut layout meant for the public facing parts of the site. """

    def __init__(self, model, request):
        super().__init__(model, request)

        # always include the common js files
        self.request.include('common')

        if self.request.is_logged_in:
            self.request.include('sortable')

    @cached_property
    def breadcrumbs(self):
        """ Returns the breadcrumbs for the current page. """
        return [Link(_("Homepage"), self.homepage_url)]

    @cached_property
    def root_pages(self):
        query = PageCollection(self.app.session()).query(ordered=False)
        query = query.order_by(desc(Page.type), Page.order)
        query = query.filter(Page.parent_id == None)

        return self.request.exclude_invisible(query.all())

    @cached_property
    def top_navigation(self):
        return tuple(
            Link(r.title, self.request.link(r)) for r in self.root_pages
        )

    @cached_property
    def bottom_links(self):

        request = self.request

        if request.current_role == 'editor':
            return [
                Link(_('Logout'), self.logout_url),
                Link(_('User Profile'), request.link(
                    self.town, 'benutzerprofil'
                )),
                Link(_('Files'), request.link(
                    GeneralFileCollection(self.app)
                )),
                Link(_('Images'), request.link(
                    ImageFileCollection(self.app)
                )),
                Link('OneGov Cloud', 'http://www.onegovcloud.ch'),
                Link('Seantis GmbH', 'https://www.seantis.ch')
            ]
        elif request.current_role == 'admin':
            return [
                Link(_('Logout'), self.logout_url),
                Link(_('User Profile'), request.link(
                    self.town, 'benutzerprofil'
                )),
                Link(_('Files'), request.link(
                    GeneralFileCollection(self.app)
                )),
                Link(_('Images'), request.link(
                    ImageFileCollection(self.app)
                )),
                Link(_('Settings'), request.link(self.town, 'einstellungen')),
                Link('OneGov Cloud', 'http://www.onegovcloud.ch'),
                Link('Seantis GmbH', 'https://www.seantis.ch')
            ]
        else:
            return [
                Link(_('Login'), self.login_url),
                Link('OneGov Cloud', 'http://www.onegovcloud.ch'),
                Link('Seantis GmbH', 'https://www.seantis.ch')
            ]


class AdjacencyListLayout(DefaultLayout):
    """ Provides layouts for for models inheriting from
    :class:`onegov.core.orm.abstract.AdjacencyList`

    """

    @cached_property
    def sortable_url_template(self):
        return self.csrf_protected_url(
            self.request.link(PageMove.for_url_template())
        )

    def get_breadcrumbs(self, item):
        """ Yields the breadcrumbs for the given adjacency list item. """

        yield Link(_("Homepage"), self.homepage_url)

        for ancestor in item.ancestors:
            yield Link(ancestor.title, self.request.link(ancestor))

        yield Link(item.title, self.request.link(item))

    def get_sidebar(self, type=None):
        """ Yields the sidebar for the given adjacency list item. """
        query = self.model.siblings.filter(self.model.__class__.type == type)
        items = self.request.exclude_invisible(query.all())

        for item in items:
            if item != self.model:
                yield Link(item.title, self.request.link(item), model=item)
            else:
                children = (
                    Link(c.title, self.request.link(c), model=c) for c
                    in self.request.exclude_invisible(self.model.children)
                )

                yield LinkGroup(
                    title=item.title,
                    links=tuple(children),
                    model=item
                )


class PageLayout(AdjacencyListLayout):

    @cached_property
    def breadcrumbs(self):
        return tuple(self.get_breadcrumbs(self.model))

    @cached_property
    def sidebar_links(self):
        return tuple(self.get_sidebar(type='topic'))


class NewsLayout(AdjacencyListLayout):

    @cached_property
    def breadcrumbs(self):
        return tuple(self.get_breadcrumbs(self.model))


class EditorLayout(AdjacencyListLayout):

    def __init__(self, model, request, site_title):
        super().__init__(model, request)
        self.site_title = site_title
        self.include_editor()

    @cached_property
    def breadcrumbs(self):
        links = list(self.get_breadcrumbs(self.model.page))
        links.append(Link(self.site_title, url='#'))

        return links


class FormEditorLayout(DefaultLayout):

    def __init__(self, model, request):
        super().__init__(model, request)
        self.include_editor()
        self.include_code_editor()


class FormSubmissionLayout(DefaultLayout):

    def __init__(self, model, request, title=None):
        super().__init__(model, request)
        self.title = title or self.form.title

    @cached_property
    def form(self):
        if hasattr(self.model, 'form'):
            return self.model.form
        else:
            return self.model

    @cached_property
    def breadcrumbs(self):
        collection = FormCollection(self.request.app.session())

        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Forms"), self.request.link(collection)),
            Link(self.title, '#')
        ]

    @cached_property
    def editbar_links(self):

        if not self.request.is_logged_in:
            return

        # only show the edit bar links if the site is the base of the form
        # -> if the user already entered some form data remove the edit bar
        # because it makes it seem like it's there to edit the submission,
        # not the actual form
        if hasattr(self.model, 'form'):
            return

        collection = FormCollection(self.request.app.session())

        edit_link = Link(
            text=_("Edit"),
            url=self.request.link(self.form, name='bearbeiten'),
            classes=('edit-link', )
        )

        if self.form.has_submissions(with_state='complete'):
            delete_link = DeleteLink(
                text=_("Delete"),
                url=self.request.link(self.form),
                confirm=_("This form can't be deleted."),
                extra_information=_(
                    "The are submissions associated with the form. "
                    "Those need to be removed first."
                )
            )

        else:
            delete_link = DeleteLink(
                text=_("Delete"),
                url=self.request.link(self.form),
                confirm=_("Do you really want to delete this form?"),
                yes_button_text=_("Delete form"),
                redirect_after=self.request.link(collection)
            )

        return [
            edit_link, delete_link
        ]


class FormCollectionLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Forms"), '#')
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            return [
                LinkGroup(
                    title=_("Add"),
                    links=[
                        Link(
                            text=_("Form"),
                            url=self.request.link(
                                self.model,
                                name='neu'
                            ),
                            classes=('new-form', )
                        )
                    ]
                ),
            ]


class PersonCollectionLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("People"), '#')
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            return [
                LinkGroup(
                    title=_("Add"),
                    links=[
                        Link(
                            text=_("Person"),
                            url=self.request.link(
                                self.model,
                                name='neu'
                            ),
                            classes=('new-person', )
                        )
                    ]
                ),
            ]


class PersonLayout(DefaultLayout):

    @cached_property
    def collection(self):
        return PersonCollection(self.request.app.session())

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("People"), self.request.link(self.collection)),
            Link(_(self.model.title), self.request.link(self.model))
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            return [
                Link(
                    text=_("Edit"),
                    url=self.request.link(self.model, 'bearbeiten'),
                    classes=('edit-link', )
                ),
                DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model),
                    confirm=_(
                        "Do you really want to delete this person?"),
                    yes_button_text=_("Delete person"),
                    redirect_after=self.request.link(self.collection)
                )
            ]


class TicketsLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Tickets"), '#')
        ]


class TicketLayout(DefaultLayout):

    @cached_property
    def collection(self):
        return TicketCollection(self.request.app.session())

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Tickets"), self.request.link(self.collection)),
            Link(self.model.number, '#')
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:

            # only show the model related links when the ticket is pending
            if self.model.state == 'pending':
                links = self.model.handler.get_links(self.request)
            else:
                links = []

            if self.model.state == 'open':
                links.append(Link(
                    text=_("Accept ticket"),
                    url=self.request.link(self.model, 'accept'),
                    classes=('ticket-button', 'ticket-accept'),
                ))

            elif self.model.state == 'pending':
                links.append(Link(
                    text=_("Close ticket"),
                    url=self.request.link(self.model, 'close'),
                    classes=('ticket-button', 'ticket-close'),
                ))

            elif self.model.state == 'closed':
                links.append(Link(
                    text=_("Reopen ticket"),
                    url=self.request.link(self.model, 'reopen'),
                    classes=('ticket-button', 'ticket-reopen'),
                ))

            return links


class ResourcesLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Reservations"), self.request.link(self.model))
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            return [
                LinkGroup(
                    title=_("Add"),
                    links=[
                        Link(
                            text=_("Room"),
                            url=self.request.link(
                                self.model,
                                name='neuer-raum'
                            ),
                            classes=('new-room', )
                        ),
                        Link(
                            text=_("Daypass"),
                            url=self.request.link(
                                self.model,
                                name='neue-tageskarte'
                            ),
                            classes=('new-daypass', )
                        )
                    ]
                ),
            ]


class ResourceLayout(DefaultLayout):

    def __init__(self, model, request):
        super().__init__(model, request)

        self.request.include('fullcalendar')

    @cached_property
    def collection(self):
        return ResourceCollection(self.request.app.libres_context)

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Reservations"), self.request.link(self.collection)),
            Link(_(self.model.title), self.request.link(self.model))
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            if self.model.deletable:
                delete_link = DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model),
                    confirm=_("Do you really want to delete this resource?"),
                    yes_button_text=_("Delete resource"),
                    redirect_after=self.request.link(self.collection)
                )

            else:
                delete_link = DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model),
                    confirm=_("This resource can't be deleted."),
                    extra_information=_(
                        "There are existing reservations associated "
                        "with this resource"
                    )
                )
            return [
                Link(
                    text=_("Edit"),
                    url=self.request.link(self.model, 'bearbeiten'),
                    classes=('edit-link', )
                ),
                delete_link,
                Link(
                    text=_("Clean up"),
                    url=self.request.link(self.model, 'cleanup'),
                    classes=('cleanup-link', 'calendar-dependent')
                ),
                Link(
                    text=_("Occupancy"),
                    url=self.request.link(self.model, 'belegung'),
                    classes=('occupancy-link', 'calendar-dependent')
                ),
                Link(
                    text=_("Export"),
                    url=self.request.link(self.model, 'export'),
                    classes=('export-link', 'calendar-dependent')
                )
            ]


class ReservationLayout(ResourceLayout):
    editbar_links = None


class AllocationEditFormLayout(DefaultLayout):
    """ Same as the resource layout, but with different editbar links, because
    there's not really an allocation view, but there are allocation forms.

    """

    @cached_property
    def collection(self):
        return ResourceCollection(self.request.app.libres_context)

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Reservations"), self.request.link(self.collection)),
            Link(_("Edit allocation"), '#')
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            if self.model.availability == 100.0:
                yield DeleteLink(
                    _("Delete"),
                    self.request.link(self.model),
                    confirm=_("Do you really want to delete this allocation?"),
                    yes_button_text=_("Delete allocation"),
                    redirect_after=self.request.link(self.collection)
                )
            else:
                yield DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model),
                    confirm=_("This resource can't be deleted."),
                    extra_information=_(
                        "There are existing reservations associated "
                        "with this resource"
                    ),
                    redirect_after=self.request.link(self.collection)
                )


class EventBaseLayout(DefaultLayout):

    event_format = 'EEEE, d. MMMM YYYY, HH:mm'

    def format_recurrence(self, recurrence):
        """ Returns a human readable version of an RRULE used by us. """

        WEEKDAYS = (_("Mo"), _("Tu"), _("We"), _("Th"), _("Fr"), _("Sa"),
                    _("Su"))

        if recurrence:
            rule = rrule.rrulestr(recurrence)
            if rule._freq == rrule.WEEKLY:
                return _(
                    "Every ${days} until ${end}",
                    mapping={
                        'days': ', '.join((
                            self.request.translate(WEEKDAYS[day])
                            for day in rule._byweekday
                        )),
                        'end': rule._until.date().strftime('%d.%m.%Y')
                    }
                )

        return ''

    def event_deletable(self, event):
        tickets = TicketCollection(self.app.session())
        ticket = tickets.by_handler_id(event.id.hex)
        return False if ticket else True


class OccurrencesLayout(EventBaseLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Events"), self.request.link(self.model))
        ]


class OccurrenceLayout(EventBaseLayout):

    @cached_property
    def collection(self):
        return OccurrenceCollection(self.request.app.session())

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Events"), self.request.link(self.collection)),
            Link(self.model.title, '#')
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            edit_url = URL(self.request.link(self.model.event, 'bearbeiten'))
            edit_url = edit_url.query_param(
                'return-to', self.request.link(self.model.event)
            )
            edit_link = Link(
                text=_("Edit"),
                url=edit_url.as_string(),
                classes=('edit-link', )
            )

            if self.event_deletable(self.model.event):
                delete_link = DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model.event),
                    confirm=_("Do you really want to delete this event?"),
                    yes_button_text=_("Delete event"),
                    redirect_after=self.events_url
                )
            else:
                delete_link = DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model.event),
                    confirm=_("This event can't be deleted."),
                    extra_information=_(
                        "To remove this event, go to the ticket and reject it."
                    )
                )

            return [edit_link, delete_link]


class EventLayout(EventBaseLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Events"), self.events_url),
            Link(self.model.title, self.request.link(self.model)),
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            edit_link = Link(
                text=_("Edit"),
                url=self.request.link(self.model, 'bearbeiten'),
                classes=('edit-link', )
            )
            if self.event_deletable(self.model):
                delete_link = DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model),
                    confirm=_("Do you really want to delete this event?"),
                    yes_button_text=_("Delete event"),
                    redirect_after=self.events_url
                )
            else:
                delete_link = DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model),
                    confirm=_("This event can't be deleted."),
                    extra_information=_(
                        "To remove this event, go to the ticket and reject it."
                    )
                )

            return [edit_link, delete_link]


class NewsletterLayout(DefaultLayout):

    @cached_property
    def collection(self):
        return NewsletterCollection(self.app.session())

    @cached_property
    def recipients(self):
        return RecipientCollection(self.app.session())

    @cached_property
    def is_collection(self):
        return isinstance(self.model, NewsletterCollection)

    @cached_property
    def breadcrumbs(self):

        if self.is_collection and self.view_name == 'neu':
            return [
                Link(_("Homepage"), self.homepage_url),
                Link(_("Newsletter"), self.request.link(self.collection)),
                Link(_("New"), '#')
            ]
        elif self.is_collection:
            return [
                Link(_("Homepage"), self.homepage_url),
                Link(_("Newsletter"), '#')
            ]
        else:
            return [
                Link(_("Homepage"), self.homepage_url),
                Link(_("Newsletter"), self.request.link(self.collection)),
                Link(self.model.title, '#')
            ]

    @cached_property
    def editbar_links(self):
        if not self.request.is_logged_in:
            return

        if self.is_collection:
            return [
                Link(
                    text=_("Subscribers"),
                    url=self.request.link(self.recipients),
                    classes=('manage-subscribers', )
                ),
                LinkGroup(
                    title=_("Add"),
                    links=[
                        Link(
                            text=_("Newsletter"),
                            url=self.request.link(
                                NewsletterCollection(self.app.session()),
                                name='neu'
                            ),
                            classes=('new-newsletter', )
                        ),
                    ]
                ),
            ]
        else:
            return [
                Link(
                    text=_("Send"),
                    url=self.request.link(self.model, 'senden'),
                    classes=('send-link', )
                ),
                Link(
                    text=_("Edit"),
                    url=self.request.link(self.model, 'bearbeiten'),
                    classes=('edit-link', )
                ),
                DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model),
                    confirm=_(
                        'Do you really want to delete "{}"?'.format(
                            self.model.title
                        )),
                    yes_button_text=_("Delete newsletter"),
                    redirect_after=self.request.link(self.collection)
                )
            ]


class RecipientLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Newsletter"), self.request.link(
                NewsletterCollection(self.app.session())
            )),
            Link(_("Subscribers"), '#')
        ]


class ImageSetCollectionLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Photo Albums"), self.request.link(self.model))
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            return [
                Link(
                    text=_("Manage images"),
                    url=self.request.link(ImageFileCollection(self.app)),
                    classes=('upload', )
                ),
                LinkGroup(
                    title=_("Add"),
                    links=[
                        Link(
                            text=_("Photo Album"),
                            url=self.request.link(
                                self.model,
                                name='neu'
                            ),
                            classes=('new-photo-album', )
                        )
                    ]
                ),
            ]


class ImageSetLayout(DefaultLayout):

    @property
    def collection(self):
        return ImageSetCollection(self.request.app.session())

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Photo Albums"), self.request.link(self.collection)),
            Link(self.model.title, self.request.link(self.model))
        ]

    @cached_property
    def editbar_links(self):
        if self.request.is_logged_in:
            return [
                Link(
                    text=_("Choose images"),
                    url=self.request.link(self.model, 'auswahl'),
                    classes=('select', )
                ),
                Link(
                    text=_("Edit"),
                    url=self.request.link(
                        self.model,
                        name='bearbeiten'
                    ),
                    classes=('edit-link', )
                ),
                DeleteLink(
                    text=_("Delete"),
                    url=self.request.link(self.model),
                    confirm=_(
                        'Do you really want to delete "{}"?'.format(
                            self.model.title
                        )),
                    yes_button_text=_("Delete photo album"),
                    redirect_after=self.request.link(self.collection)
                )
            ]



from onegov.core.security import Private
from onegov.town import TownApp
from onegov.town.models import PageMove


@TownApp.view(model=PageMove, permission=Private, request_method='PUT')
def move_page(self, request):
    request.assert_valid_csrf_token()
    self.execute()



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
import lxml
import re

from base64 import b64decode
from json import loads, dumps
from onegov.core.orm import find_models
from onegov.core.orm.abstract import sort_siblings
from onegov.core.orm.mixins import ContentMixin
from onegov.core.upgrade import upgrade_task
from onegov.form import FormCollection, FormDefinition
from onegov.libres import ResourceCollection
from onegov.page import PageCollection
from onegov.people import Person
from onegov.town import TownApp
from onegov.town.initial_content import add_resources
from onegov.town.models.extensions import ContactExtension
from onegov.town.models.file import (
    GeneralFileCollection,
    ImageFile,
    ImageFileCollection
)
from pytz import utc


@upgrade_task('Add initial libres resources')
def add_initial_libres_resources(context):
    if isinstance(context.app, TownApp):
        add_resources(context.app.libres_context)


@upgrade_task('Add order to all pages')
def add_order_to_all_pages(context):
    pages = PageCollection(context.session)
    processed_siblings = set()

    for page in pages.query(ordered=False):
        siblings = page.siblings.all()

        ids = {sibling.id for sibling in siblings}
        if ids <= processed_siblings:
            continue

        sort_siblings(siblings, key=PageCollection.sort_key)

        processed_siblings.update(ids)


@upgrade_task('Update contact html')
def change_contact_html(context):
    items = []

    items.extend(PageCollection(context.session).query().all())
    items.extend(FormCollection(context.session).definitions.query().all())

    if isinstance(context.app, TownApp):
        items.extend(
            ResourceCollection(context.app.libres_context).query().all()
        )

    for item in items:
        if not isinstance(item, ContactExtension):
            continue
        if not item.contact:
            continue

        # forces a re-render of the contact html
        item.contact = item.contact


@upgrade_task('Convert builtin forms into custom forms')
def convert_builtin_forms(context):
    forms = FormCollection(context.session)

    query = forms.definitions.query()
    query = query.filter(FormDefinition.type == 'builtin')
    query.update({FormDefinition.type: 'custom'})


@upgrade_task('Migrate images/files to onegov.file')
def migrate_images_file_to_onegov_file(context):

    request, app = context.request, context.app

    mapping = {
        'images': ImageFileCollection(context.app.session()),
        'files': GeneralFileCollection(context.app.session())
    }

    for folder, collection in mapping.items():
        if not app.filestorage.isdir(folder):
            continue

        storage = app.filestorage.opendir(folder)
        rewrite = []

        for filename in storage.listdir(files_only=True):
            if filename.endswith('.r'):
                continue

            if '-' in filename:
                name = filename.split('-')[0].strip()
                name = b64decode(name).decode('utf-8')
            else:
                name = filename

            with storage.open(filename, mode='rb') as f:
                try:
                    new_file = collection.add(name, content=f)
                except FileExistsError as e:
                    new_file = e.args[0]

            with storage.open(filename + '.r', mode='w') as f:
                f.write(new_file.id)

            fileinfo = storage.getinfo(filename)
            new_file.created = fileinfo['created_time'].replace(tzinfo=utc)
            new_file.modified = fileinfo['modified_time'].replace(tzinfo=utc)

            old_url = request.filestorage_link('/'.join((folder, filename)))
            new_url = request.link(new_file)

            # only work with the last two parts of the url, as the front with
            # the domain may be different during the upgrade
            old_url = '/'.join(old_url.split('/')[-2:])
            new_url = '/'.join(new_url.split('/')[-2:])

            # adjust the prefix of the paths which was different when the
            # paths for it were still configured by onegov.town
            old_url = old_url.replace('images/', 'bild/')
            old_url = old_url.replace('files/', 'datei/')

            rewrite.append((old_url, new_url))

        # update all records with a meta/content mixin
        def is_match(cls):
            return issubclass(cls, ContentMixin)

        for base in app.session_manager.bases:
            for cls in find_models(base, is_match):
                for item in app.session().query(cls).all():
                    meta, content = dumps(item.meta), dumps(item.content)

                    for old, new in rewrite:
                        meta = meta.replace(old, new)
                        content = content.replace(old, new)

                    item.meta = loads(meta)
                    item.content = loads(content)

        # update the theme options
        town = app.load_town()
        theme_options = dumps(town.theme_options)

        for old, new in rewrite:
            theme_options = theme_options.replace(old, new)

        town.theme_options = loads(theme_options)

        # update the people images
        for person in app.session().query(Person).all():
            for old, new in rewrite:
                person.picture_url = person.picture_url.replace(old, new)


@upgrade_task('Migrate alt text',
              requires='onegov.town:Migrate images/files to onegov.file')
def migrate_alt_text(context):

    app = context.app
    url_expression = re.compile(r'^.*/storage/([a-z0-9]{64})/?.*$')

    # find all possible alt texts and transfer them to onegov.file
    def is_match(cls):
        return issubclass(cls, ContentMixin)

    items = []

    for base in app.session_manager.bases:
        for cls in find_models(base, is_match):
            for item in app.session().query(cls).all():
                if getattr(item, 'text', None):
                    items.append(item)

    alts = {}

    for item in items:
        html = lxml.html.fromstring(item.text)
        for img in html.xpath('//img'):
            src = (img.get('src') or '').strip()

            if not src:
                continue

            # check if it's one of the internal images
            match = url_expression.match(src)

            if not match:
                continue

            # remember the alt text for each image
            if img.get('alt'):
                fid = match.group(1)
                alts[fid] = img.get('alt')
                img.set('alt', '')

            # load the alt text lazily from now on (might be added later)
            img.set('class', 'lazyload-alt')

        item.text = lxml.html.tostring(html).decode('utf-8')

    if not alts:
        return

    query = ImageFileCollection(app.session()).query()
    query = query.filter(ImageFile.id.in_(alts.keys()))
    files = {f.id: f for f in query.all()}

    for fid in alts:
        if fid in files:
            files[fid].note = alts[fid]



from onegov.form import Form
from onegov.town import _
from wtforms import StringField, validators


class SignupForm(Form):

    address = StringField(
        label=_("E-Mail Address"),
        validators=[validators.InputRequired(), validators.Email()]
    )



from onegov.form.models import FormDefinition
from onegov.libres.models import Resource
from onegov.town.models import News, Topic, ImageSet
from sqlalchemy.orm import defer


class SiteCollection(object):

    def __init__(self, session):
        self.session = session

    def get(self):
        # get the topics
        topics = self.session.query(Topic)
        topics = topics.options(defer(Topic.meta))
        topics = topics.options(defer(Topic.content))
        topics = topics.options(defer(Topic.order))
        topics = topics.order_by(Topic.title)

        # get the news
        news = self.session.query(News)
        news = news.options(defer(News.meta))
        news = news.options(defer(News.content))
        news = news.options(defer(News.order))
        news = news.order_by(News.title)

        # get the forms
        forms = self.session.query(FormDefinition)
        forms = forms.options(defer(FormDefinition.definition))
        forms = forms.options(defer(FormDefinition.checksum))
        forms = forms.order_by(FormDefinition.title)

        # get the resources
        resources = self.session.query(Resource)
        resources = resources.options(defer(Resource.timezone))
        resources = resources.order_by(Resource.title)

        # get the foto albums
        imagesets = self.session.query(ImageSet)
        imagesets = imagesets.options(defer(ImageSet.meta))
        imagesets = imagesets.options(defer(ImageSet.content))
        imagesets = imagesets.order_by(ImageSet.title)

        return {
            'topics': topics.all(),
            'news': news.all(),
            'forms': forms.all(),
            'resources': resources.all(),
            'imagesets': imagesets.all(),
        }



from onegov.town import utils


def test_annotate_html():
    html = "<p><img/></p><p></p>"
    assert utils.annotate_html(html) == (
        '<p class="has-img"><img class="lazyload-alt"></p><p></p>'
    )

    html = "<p class='x'><img/></p><p></p>"
    assert utils.annotate_html(html) == (
        '<p class="x has-img"><img class="lazyload-alt"></p><p></p>'
    )

    html = "<strong></strong>"
    assert utils.annotate_html(html) == html

    html = '<a href="http://www.seantis.ch"></a>'
    assert utils.annotate_html(html) == html

    html = 'no html'
    assert utils.annotate_html(html) == 'no html'

    html = '<p><a href="http://www.seantis.ch"></a></p>'
    assert utils.annotate_html(html) == html

    html = (
        '<p><a href="https://www.youtube.com/watch?v=dQw4w9WgXcQ"></a></p>'
    )
    assert '<p class="has-video">' in utils.annotate_html(html)
    assert 'class="has-video"></a>' in utils.annotate_html(html)

    html = (
        '<p>'
        '<a href="https://www.youtube.com/watch?v=dQw4w9WgXcQ"></a>'
        '<img />'
        '</p>'
    )
    assert '<p class="has-img has-video">' in utils.annotate_html(html)
    assert 'class="has-video"></a>' in utils.annotate_html(html)



from onegov.form.models import FormDefinition
from onegov.town.models.extensions import (
    HiddenFromPublicExtension,
    ContactExtension,
    PersonLinkExtension,
    CoordinatesExtension
)


class BuiltinFormDefinition(FormDefinition, HiddenFromPublicExtension,
                            ContactExtension, PersonLinkExtension,
                            CoordinatesExtension):
    __mapper_args__ = {'polymorphic_identity': 'builtin'}

    es_type_name = 'builtin_forms'

    @property
    def es_public(self):
        return not self.is_hidden_from_public


class CustomFormDefinition(FormDefinition, HiddenFromPublicExtension,
                           ContactExtension, PersonLinkExtension,
                           CoordinatesExtension):
    __mapper_args__ = {'polymorphic_identity': 'custom'}

    es_type_name = 'custom_forms'

    @property
    def es_public(self):
        return not self.is_hidden_from_public



""" Contains the application which builds on onegov.core and uses
more.chameleon.

It's possible that the chameleon integration is moved to onegov.core in the
future, but for now it is assumed that different applications may want to
use different templating languages.

"""

from collections import defaultdict
from contextlib import contextmanager
from onegov.core import Framework, utils
from onegov.file import DepotApp
from onegov.gis import MapboxApp
from onegov.libres import LibresIntegration
from onegov.page import PageCollection
from onegov.search import ElasticsearchApp
from onegov.ticket import TicketCollection
from onegov.town.models import Town, Topic
from onegov.town.theme import TownTheme
from sqlalchemy.orm.attributes import flag_modified


class TownApp(Framework, LibresIntegration, ElasticsearchApp, MapboxApp,
              DepotApp):
    """ The town application. Include this in your onegov.yml to serve it
    with onegov-server.

    """

    serve_static_files = True

    def is_allowed_application_id(self, application_id):
        """ Stops onegov.server from ever passing the request to the town
        application, if the schema does not exist. This way we can host
        onegov.town in a way that allows all requests to *.onegovcloud.ch.

        If the schema for ``newyork.onegovcloud.ch`` exists, the request is
        handled. If the schema does not exist, the request is not handled.

        Here we basically decide if a town exists or not.

        """
        schema = self.namespace + '-' + application_id

        if schema in self.known_schemas:
            return True

        # block invalid schemas from ever being checked
        if not self.session_manager.is_valid_schema(schema):
            return False

        # if the schema exists, remember it
        if self.session_manager.is_schema_found_on_database(schema):
            self.known_schemas.add(schema)

            return True

        return False

    @property
    def town(self):
        """ Returns the cached version of the town. Since the town rarely
        ever changes, it makes sense to not hit the database for it every
        time.

        As a consequence, changes to the town object are not propagated,
        unless you use :meth:`update_town` or use the ORM directly.

        """
        town = self.cache.get_or_create(
            'town',
            creator=self.load_town,
            should_cache_fn=lambda town: town is not None
        )

        if town is not None:
            return self.session().merge(town, load=False)

    def load_town(self):
        """ Loads the town from the SQL database. """
        return self.session().query(Town).first()

    @contextmanager
    def update_town(self):
        """ Yields the current town for an update. Use this instead of
        updating the town directly, because caching is involved. It's rather
        easy to otherwise update it wrongly.

        Example::
            with app.update_town() as town:
                town.name = 'New Name'

        """

        session = self.session()

        town = session.merge(self.town)
        yield town

        # nested entries in the meta json field are not detected as modified
        flag_modified(town, 'meta')

        session.flush()

        self.cache.delete('town')

    @property
    def ticket_count(self):
        return self.cache.get_or_create('ticket_count', self.load_ticket_count)

    def load_ticket_count(self):
        return TicketCollection(self.session()).get_count()

    def update_ticket_count(self):
        return self.cache.delete('ticket_count')

    @property
    def homepage_pages(self):
        return self.cache.get_or_create(
            'homepage_pages', self.load_homepage_pages)

    def load_homepage_pages(self):
        pages = PageCollection(self.session()).query()
        pages = pages.filter(Topic.type == 'topic')

        # XXX use JSON/JSONB for this (the attribute is not there if it's
        # false, so this is not too bad speed-wise but it's still awful)
        pages = pages.filter(Topic.meta.contains(
            'is_visible_on_homepage'
        ))

        result = defaultdict(list)
        for page in pages.all():
            if page.is_visible_on_homepage:
                result[page.root.id].append(page)

        for key in result:
            result[key] = list(sorted(
                result[key],
                key=lambda p: utils.normalize_for_url(p.title)
            ))

        return result

    def update_homepage_pages(self):
        return self.cache.delete('homepage_pages')

    def send_email(self, **kwargs):
        """ Wraps :meth:`onegov.core.framework.Framework.send_email`, setting
        the reply_to address by using the reply address from the town
        settings.

        """

        assert 'reply_to' in self.town.meta

        reply_to = "{} <{}>".format(self.town.name, self.town.meta['reply_to'])

        return super().send_email(reply_to=reply_to, **kwargs)

    def configure_application(self, **cfg):
        super().configure_application(**cfg)

        if self.has_database_connection:
            schema_prefix = self.namespace + '-'

            self.known_schemas = set(
                s for s in self.session_manager.list_schemas()
                if s.startswith(schema_prefix)
            )

    @property
    def theme_options(self):
        return self.town.theme_options or {}


@TownApp.template_directory()
def get_template_directory():
    return 'templates'


@TownApp.setting(section='core', name='theme')
def get_theme():
    return TownTheme()


@TownApp.setting(section='i18n', name='localedirs')
def get_i18n_localedirs():
    return [
        utils.module_path('onegov.town', 'locale'),
        utils.module_path('onegov.form', 'locale'),
        utils.module_path('onegov.user', 'locale')
    ]


@TownApp.setting(section='i18n', name='default_locale')
def get_i18n_default_locale():
    return 'de_CH'


@TownApp.webasset_path()
def get_shared_assets_path():
    return utils.module_path('onegov.shared', 'assets/js')


@TownApp.webasset_path()
def get_js_path():
    return 'assets/js'


@TownApp.webasset_path()
def get_css_path():
    return 'assets/css'


@TownApp.webasset_output()
def get_webasset_output():
    return 'assets/bundles'


@TownApp.webasset('sortable')
def get_sortable_asset():
    yield 'sortable.js'
    yield 'sortable_custom.js'


@TownApp.webasset('fullcalendar')
def get_fullcalendar_asset():
    yield 'fullcalendar.css'
    yield 'moment.js'
    yield 'moment.de.js'
    yield 'fullcalendar.js'
    yield 'fullcalendar.de.js'
    yield 'reservationcalendar.jsx'
    yield 'reservationcalendar_custom.js'


@TownApp.webasset('check_contrast')
def get_check_contrast_asset():
    yield 'check_contrast.js'


@TownApp.webasset('check_password')
def get_check_password_asset():
    yield 'zxcvbn.js'
    yield 'check_password.js'


@TownApp.webasset('code_editor')
def get_code_editor_asset():
    yield 'ace.js'
    yield 'ace-mode-form.js'
    yield 'ace-theme-tomorrow.js'
    yield 'code_editor.js'


@TownApp.webasset('editor')
def get_editor_asset():
    yield 'bufferbuttons.js'
    yield 'definedlinks.js'
    yield 'filemanager.js'
    yield 'imagemanager.js'
    yield 'redactor.de.js'
    yield 'input_with_button.js'
    yield 'editor.js'


# do NOT minify the redactor, or the copyright notice goes away, which
# is something we are not allowed to do per our license
@TownApp.webasset('redactor', filters={'js': None})
def get_redactor_asset():
    yield 'redactor.min.js'
    yield 'redactor.css'


@TownApp.webasset('dropzone')
def get_dropzone_asset():
    yield 'dropzone.js'


@TownApp.webasset('editalttext')
def get_editalttext_asset():
    yield 'editalttext.js'


@TownApp.webasset('common')
def get_common_asset():
    yield 'jquery.datetimepicker.css'
    yield 'locale.js'
    yield 'modernizr.js'
    yield 'jquery.js'
    yield 'fastclick.js'
    yield 'foundation.js'
    yield 'intercooler.js'
    yield 'underscore.js'
    yield 'react.js'
    yield 'form_dependencies.js'
    yield 'confirm.jsx'
    yield 'typeahead.jsx'
    yield 'leaflet'
    yield 'jquery.datetimepicker.js'
    yield 'jquery.mousewheel.js'
    yield 'jquery.popupoverlay.js'
    yield 'jquery.load.js'
    yield 'videoframe.js'
    yield 'datetimepicker.js'
    yield 'url.js'
    yield 'date-range-selector.js'
    yield 'lazyalttext.js'
    yield 'lazysizes.js'
    yield 'common.js'



import morepath
import onegov.core
import onegov.town
import pytest
import tempfile
import transaction
import shutil

from onegov.core.crypto import hash_password
from onegov.core.utils import Bunch, scan_morepath_modules
from onegov.town.initial_content import (
    add_initial_content, builtin_form_definitions
)
from onegov.town.models import Town
from onegov.user import User
from uuid import uuid4


@pytest.yield_fixture(scope='session')
def handlers():
    yield onegov.ticket.handlers
    onegov.ticket.handlers.registry = {}


@pytest.fixture(scope='session')
def town_password():
    # only hash the password for the test users once per test session
    return hash_password('hunter2')


@pytest.yield_fixture(scope='session')
def filestorage():
    directory = tempfile.mkdtemp()
    yield directory
    shutil.rmtree(directory)


@pytest.yield_fixture(scope='session')
def form_definitions():
    yield list(builtin_form_definitions())


@pytest.yield_fixture(scope='function')
def town_app(postgres_dsn, filestorage, town_password, smtp,
             form_definitions):
    yield new_town_app(
        postgres_dsn, filestorage, town_password, smtp, form_definitions
    )


@pytest.yield_fixture(scope='function')
def es_town_app(postgres_dsn, filestorage, town_password, smtp,
                form_definitions, es_url):
    yield new_town_app(
        postgres_dsn, filestorage, town_password, smtp, form_definitions,
        es_url
    )


def new_town_app(postgres_dsn, filestorage, town_password, smtp,
                 form_definitions, es_url=None):

    scan_morepath_modules(onegov.town.TownApp)
    morepath.commit(onegov.town.TownApp)

    app = onegov.town.TownApp()
    app.namespace = 'test_' + uuid4().hex
    app.configure_application(
        dsn=postgres_dsn,
        filestorage='fs.osfs.OSFS',
        filestorage_options={
            'root_path': filestorage,
            'create': True
        },
        depot_backend='depot.io.memory.MemoryFileStorage',
        identity_secure=False,
        disable_memcached=True,
        enable_elasticsearch=es_url and True or False,
        elasticsearch_hosts=[es_url]
    )
    app.set_application_id(app.namespace + '/' + 'test')
    app.bind_depot()
    add_initial_content(
        app.libres_registry,
        app.session_manager,
        'Govikon',
        form_definitions
    )

    # cronjobs leave lingering sessions open, in real life this is not a
    # problem, but in testing it leads to connection pool exhaustion
    app.settings.cronjobs = Bunch(enabled=False)

    session = app.session()

    town = session.query(Town).one()
    town.meta['reply_to'] = 'mails@govikon.ch'

    app.mail_host, app.mail_port = smtp.address
    app.mail_sender = 'mails@govikon.ch'
    app.mail_force_tls = False
    app.mail_username = None
    app.mail_password = None
    app.mail_use_directory = False
    app.smtp = smtp

    # usually we don't want to create the users directly, anywhere else you
    # *need* to go through the UserCollection. Here however, we can improve
    # the test speed by not hashing the password for every test.

    session.add(User(
        username='admin@example.org',
        password_hash=town_password,
        role='admin'
    ))
    session.add(User(
        username='editor@example.org',
        password_hash=town_password,
        role='editor'
    ))

    transaction.commit()
    session.close_all()

    return app



import morepath

from onegov.core.security import Public, Private
from onegov.people import Person, PersonCollection
from onegov.town import _, TownApp
from onegov.town.elements import Link
from onegov.town.forms import PersonForm
from onegov.town.layout import PersonLayout, PersonCollectionLayout
from onegov.town.models import AtoZ


@TownApp.html(model=PersonCollection, template='people.pt', permission=Public)
def view_people(self, request):

    people = self.query().order_by(Person.last_name, Person.first_name)

    class AtoZPeople(AtoZ):

        def get_title(self, item):
            return item.title

        def get_items(self):
            return people

    return {
        'title': _("People"),
        'people': AtoZPeople(request).get_items_by_letter().items(),
        'layout': PersonCollectionLayout(self, request)
    }


@TownApp.html(model=Person, template='person.pt', permission=Public)
def view_person(self, request):

    return {
        'title': self.title,
        'person': self,
        'layout': PersonLayout(self, request)
    }


@TownApp.form(model=PersonCollection, name='neu', template='form.pt',
              permission=Private, form=PersonForm)
def handle_new_person(self, request, form):

    if form.submitted(request):
        person = self.add(**form.get_useful_data())
        request.success(_("Added a new person"))

        return morepath.redirect(request.link(person))

    layout = PersonCollectionLayout(self, request)
    layout.breadcrumbs.append(Link(_("New"), '#'))
    layout.include_editor()

    return {
        'layout': layout,
        'title': _("New person"),
        'form': form
    }


@TownApp.form(model=Person, name='bearbeiten', template='form.pt',
              permission=Private, form=PersonForm)
def handle_edit_person(self, request, form):

    if form.submitted(request):
        form.populate_obj(self)
        request.success(_("Your changes were saved"))

        return morepath.redirect(request.link(self))
    else:
        form.process(obj=self)

    layout = PersonLayout(self, request)
    layout.breadcrumbs.append(Link(_("Edit"), '#'))
    layout.include_editor()

    return {
        'layout': layout,
        'title': self.title,
        'form': form
    }


@TownApp.view(model=Person, request_method='DELETE', permission=Private)
def handle_delete_person(self, request):
    PersonCollection(request.app.session()).delete(self)



""" Contains the paths to the different models served by onegov.town. """

from datetime import date
from libres.db.models import Allocation, Reservation
from onegov.event import (
    Event,
    EventCollection,
    Occurrence,
    OccurrenceCollection
)
from onegov.form import (
    FormDefinition,
    FormCollection,
    FormSubmissionFile,
    CompleteFormSubmission,
    PendingFormSubmission
)
from onegov.libres import ResourceCollection
from onegov.libres.models import Resource
from onegov.newsletter import (
    Newsletter,
    NewsletterCollection,
    Subscription,
    RecipientCollection,
)
from onegov.town.app import TownApp
from onegov.town.converters import extended_date_converter
from onegov.town.models import (
    AtoZPages,
    Clipboard,
    Editor,
    FormPersonMove,
    GeneralFileCollection,
    ImageFileCollection,
    ImageSet,
    ImageSetCollection,
    LegacyFile,
    LegacyFileCollection,
    LegacyImage,
    LegacyImageCollection,
    News,
    PageMove,
    PagePersonMove,
    ResourcePersonMove,
    Search,
    SiteCollection,
    Topic,
    Town,
)
from onegov.page import PageCollection
from onegov.people import Person, PersonCollection
from onegov.ticket import Ticket, TicketCollection
from onegov.user import Auth
from webob import exc


@TownApp.path(model=Town, path='/')
def get_town(app):
    return app.town


@TownApp.path(model=Auth, path='/auth')
def get_auth(app, to='/'):
    return Auth.from_app(app, to)


@TownApp.path(model=Topic, path='/themen', absorb=True)
def get_topic(app, absorb):
    return PageCollection(app.session()).by_path(absorb, ensure_type='topic')


@TownApp.path(model=News, path='/aktuelles', absorb=True)
def get_news(app, absorb):
    absorb = '/{}/{}'.format('aktuelles', absorb)
    return PageCollection(app.session()).by_path(absorb, ensure_type='news')


@TownApp.path(model=GeneralFileCollection, path='/dateien')
def get_files(app):
    return GeneralFileCollection(app.session())


@TownApp.path(model=ImageFileCollection, path='/bilder')
def get_images(app):
    return ImageFileCollection(app.session())


@TownApp.path(model=FormCollection, path='/formulare')
def get_forms(app):
    return FormCollection(app.session())


@TownApp.path(model=FormDefinition, path='/formular/{name}')
def get_form(app, name):
    return FormCollection(app.session()).definitions.by_name(name)


@TownApp.path(model=PendingFormSubmission, path='/formular-eingabe/{id}')
def get_pending_form_submission(app, id):
    return FormCollection(app.session()).submissions.by_id(
        id, state='pending', current_only=True)


@TownApp.path(model=CompleteFormSubmission, path='/formular-eingang/{id}')
def get_complete_form_submission(app, id):
    return FormCollection(app.session()).submissions.by_id(
        id, state='complete', current_only=False)


@TownApp.path(model=FormSubmissionFile, path='/formular-datei/{id}')
def get_form_submission_file(app, id):
    return FormCollection(app.session()).submissions.file_by_id(id)


@TownApp.path(model=Editor, path='/editor/{action}/{trait}/{page_id}')
def get_editor(app, action, trait, page_id):
    if not Editor.is_supported_action(action):
        return None

    page = PageCollection(app.session()).by_id(page_id)

    if not page.is_supported_trait(trait):
        return None

    if page is not None:
        return Editor(action=action, page=page, trait=trait)


@TownApp.path(model=PersonCollection, path='/personen')
def get_people(app):
    return PersonCollection(app.session())


@TownApp.path(model=Person, path='/person/{id}')
def get_person(app, id):
    return PersonCollection(app.session()).by_id(id)


@TownApp.path(model=Ticket, path='/ticket/{handler_code}/{id}')
def get_ticket(app, handler_code, id):
    return TicketCollection(app.session()).by_id(
        id, ensure_handler_code=handler_code)


@TownApp.path(model=TicketCollection, path='/tickets/{handler}/{state}')
def get_tickets(app, handler='ALL', state='open', page=0, group=None,
                extra_parameters=None):
    return TicketCollection(
        app.session(),
        handler=handler,
        state=state,
        page=page,
        group=group,
        extra_parameters=extra_parameters
    )


@TownApp.path(model=ResourceCollection, path='/ressourcen')
def get_resources(app):
    return app.libres_resources


@TownApp.path(model=Resource, path='/ressource/{name}', converters=dict(
              date=date, highlights_min=int, highlights_max=int))
def get_resource(app, name, date=None, view='agendaWeek',
                 highlights_min=None, highlights_max=None):

    resource = app.libres_resources.by_name(name)

    if resource:
        resource.date = date
        resource.highlights_min = highlights_min
        resource.highlights_max = highlights_max
        resource.view = view

    return resource


@TownApp.path(model=Allocation, path='/einteilung/{resource}/{id}')
def get_allocation(app, resource, id):
    resource = app.libres_resources.by_id(resource)

    if resource:
        allocation = resource.scheduler.allocations_by_ids((id, )).first()

        # always get the master, even if another id is requested
        return allocation and allocation.get_master()


@TownApp.path(model=Reservation, path='/reservation/{resource}/{id}')
def get_reservation(app, resource, id):
    resource = app.libres_resources.by_id(resource)

    if resource:
        query = resource.scheduler.managed_reservations()
        query = query.filter(Reservation.id == id)

        return query.first()


@TownApp.path(model=Clipboard, path='/clipboard/copy/{token}')
def get_clipboard(request, token):
    clipboard = Clipboard(request, token)

    # the url is None if the token is invalid
    if clipboard.url:
        return clipboard


@TownApp.path(model=SiteCollection, path='/sitecollection')
def get_sitecollection(app):
    return SiteCollection(app.session())


@TownApp.path(model=PageMove,
              path='/move/page/{subject_id}/{direction}/{target_id}',
              converters=dict(subject_id=int, target_id=int))
def get_page_move(app, subject_id, direction, target_id):
    if subject_id == target_id:
        raise exc.HTTPBadRequest()

    session = app.session()
    pages = PageCollection(session)

    subject = pages.by_id(subject_id)
    target = pages.by_id(target_id)

    if subject and target:
        return PageMove(session, subject, target, direction)


@TownApp.path(model=PagePersonMove,
              path='/move/page-person/{key}/{subject}/{direction}/{target}')
def get_person_move(app, key, subject, direction, target):
    if subject == target:
        raise exc.HTTPBadRequest()

    session = app.session()
    page = PageCollection(session).by_id(key)

    if page:
        return PagePersonMove(session, page, subject, target, direction)


@TownApp.path(model=FormPersonMove,
              path='/move/form-person/{key}/{subject}/{direction}/{target}')
def get_form_move(app, key, subject, direction, target):
    session = app.session()
    form = FormCollection(session).definitions.by_name(key)

    if form:
        return FormPersonMove(session, form, subject, target, direction)


@TownApp.path(
    model=ResourcePersonMove,
    path='/move/resource-person/{key}/{subject}/{direction}/{target}')
def get_resource_move(app, key, subject, direction, target):
    session = app.session()
    resource = ResourceCollection(app.libres_context).by_id(key)

    if resource:
        return ResourcePersonMove(
            session, resource, subject, target, direction)


@TownApp.path(model=OccurrenceCollection, path='/veranstaltungen',
              converters=dict(start=extended_date_converter,
                              end=extended_date_converter, tags=[]))
def get_occurrences(app, page=0, start=None, end=None, tags=None):
    return OccurrenceCollection(app.session(), page, start, end, tags)


@TownApp.path(model=Occurrence, path='/veranstaltung/{name}')
def get_occurrence(app, name):
    return OccurrenceCollection(app.session()).by_name(name)


@TownApp.path(model=EventCollection, path='/events')
def get_events(app, page=0, state='published'):
    return EventCollection(app.session(), page, state)


@TownApp.path(model=Event, path='/event/{name}')
def get_event(app, name):
    return EventCollection(app.session()).by_name(name)


@TownApp.path(model=Search, path='/suche')
def get_search(request, q='', page=0):
    return Search(request, q, page)


@TownApp.path(model=AtoZPages, path='/a-z')
def get_a_to_z(request):
    return AtoZPages(request)


@TownApp.path(model=NewsletterCollection, path='/newsletters')
def get_newsletters(app):
    return NewsletterCollection(app.session())


@TownApp.path(model=Newsletter, path='/newsletter/{name}')
def get_newsletter(app, name):
    return get_newsletters(app).by_name(name)


@TownApp.path(model=RecipientCollection, path='/abonnenten')
def get_newsletter_recipients(app):
    return RecipientCollection(app.session())


@TownApp.path(model=Subscription, path='/abonnement/{recipient_id}/{token}')
def get_subscription(app, recipient_id, token):
    recipient = RecipientCollection(app.session()).by_id(recipient_id)
    return recipient and Subscription(recipient, token)


@TownApp.path(model=LegacyFile, path='/datei/{filename}')
def get_file(app, filename):
    return LegacyFileCollection(app).get_file_by_filename(filename)


@TownApp.path(model=LegacyImage, path='/bild/{filename}')
def get_image(app, filename):
    return LegacyImageCollection(app).get_file_by_filename(filename)


@TownApp.path(model=ImageSetCollection, path='/fotoalben')
def get_image_sets(app):
    return ImageSetCollection(app.session())


@TownApp.path(model=ImageSet, path='/fotoalbum/{id}')
def get_image_set(app, id):
    return ImageSetCollection(app.session()).by_id(id)



""" Lists the custom forms. """

from onegov.core.security import Public
from onegov.form import FormCollection
from onegov.town import _
from onegov.town.app import TownApp
from onegov.town.layout import FormCollectionLayout
from unidecode import unidecode


@TownApp.html(model=FormCollection, template='forms.pt', permission=Public)
def view_form_collection(self, request):

    # XXX add collation support to the core (create collations automatically)
    forms = self.definitions.query().all()
    forms = sorted(forms, key=lambda d: unidecode(d.title))

    return {
        'layout': FormCollectionLayout(self, request),
        'title': _("Forms"),
        'forms': request.exclude_invisible(forms)
    }



import codecs
import mistune
import os
import textwrap

from datetime import datetime, timedelta
from onegov.core.utils import module_path
from onegov.event import EventCollection
from onegov.form import FormCollection
from onegov.libres import LibresIntegration, ResourceCollection
from onegov.page import PageCollection
from onegov.town.models import Town
from sedate import as_datetime


def add_initial_content(libres_registry, session_manager, town_name,
                        form_definitions=None, reply_to=None):
    """ Adds the initial content for the given town on the given session.
    All content that comes with a new town is added here.

    Note, the ``form_definitions`` parameter is used to speed up testing,
    you usually do not want to specify it.

    """

    session = session_manager.session()

    libres_context = LibresIntegration.libres_context_from_session_manager(
        libres_registry, session_manager)

    # can only be called if no town is defined yet
    assert not session.query(Town).first()

    if reply_to:
        meta = {
            'reply_to': reply_to
        }
    else:
        meta = {}

    session.add(Town(name=town_name, meta=meta))

    add_root_pages(session)
    add_builtin_forms(session, form_definitions)
    add_resources(libres_context)
    add_events(session, town_name)
    add_welcome_page(session)

    session.flush()


def add_root_pages(session):
    pages = PageCollection(session)

    pages.add_root(
        "Leben & Wohnen",
        name='leben-wohnen',
        type='topic',
        meta={'trait': 'page'}
    ),
    pages.add_root(
        "Kultur & Freizeit",
        name='kultur-freizeit',
        type='topic',
        meta={'trait': 'page'}
    ),
    pages.add_root(
        "Bildung & Gesellschaft",
        name='bildung-gesellschaft',
        type='topic',
        meta={'trait': 'page'}
    ),
    pages.add_root(
        "Gewerbe & Tourismus",
        name='gewerbe-tourismus',
        type='topic',
        meta={'trait': 'page'}
    ),
    pages.add_root(
        "Politik & Verwaltung",
        name='politik-verwaltung',
        type='topic',
        meta={'trait': 'page'}
    )
    pages.add_root(
        "Aktuelles",
        name='aktuelles',
        type='news',
        meta={'trait': 'news'}
    )


def add_welcome_page(session):

    # for now only in German, translating through gettext is not the right way
    # here, because it splits the text into fragments which results in an
    # incoherent text
    title = "Willkommen bei OneGov"

    lead = """
        Ihre neuer Online Schalter für Gemeinde kann ab sofort genutzt werden.
        Um erste Schritte, mehr Informationen und eine kurze Übersicht zu
        erhalten klicken Sie einfach auf den Titel dieser Nachricht.
    """.replace('\n', '').strip('').replace('  ', ' ')

    text = textwrap.dedent("""\
        ## Erste Schritte

        Die folgenden Vorschläge helfen Ihnen sich in der OneGov Cloud zurecht
        zu finden. Falls Sie das lieber auf eigene Faust tun, können Sie diese
        Schritte gerne überspringen.

        **Melden Sie sich an**

        Sie können sich ab sofort mit Ihrer E-Mail Adresse anmelden. Sie haben
        das Passwort vergessen? Kein Problem,
        [setzen Sie es zurück](/request-password).

        **Passen Sie das Aussehen Ihren Bedürfnissen an**

        OneGov Cloud lässt sich nach Ihrem Gusto einrichten. Sie können das
        Logo, die Bilder auf der Startseite und mehr in den
        [Einstellungen](/einstellungen) ändern.

        **Stellen Sie Ihre Gemeinde vor**

        Klicken Sie auf **Leben & Wohnen** in der Navigation und
        anschliessend auf **Hinzufügen > Thema** gleich unterhalb der
        Hauptnavigation. Sie sehen **Hinzufügen** nicht? Dann müssen Sie sich
        erst noch anmelden.

        **Füllen Sie ein Formular aus**

        Im [Online-Schalter](/formulare) gibt es eine ganze Reihe von
        Formularen die Bürger ausfüllen können. Füllen Sie ein Formular aus
        und ein Ticket wird geöffnet.

        In den [Tickets](/tickets/ALL/open?page=0) können Sie das Ticket
        anschliessend bearbeiten. Übrigens: Sie können jederzeit eigene
        Formulare hinzufügen. Die mitgelieferten Formulare sind lediglich
        Vorschläge.

        ## Über Ihre OneGov Cloud Instanz

        Es gibt noch viel mehr zu entdecken - wir möchten an dieser Stelle
        nicht schon alles verraten.

        Ihre OneGov Cloud Instanz läuft auf einer Testumgebung, welche
        frühestens nach zwei Wochen verfällt. Sie können OneGov Cloud also
        in aller Ruhe testen.

        Zwar ist die Cloud für alle Personen zugänglich, wir stellen aber
        sicher dass sie nicht von Suchmaschinen gefunden wird. Wir
        möchten ja nicht das Ihre Bürger auf der falschen Seite landen.

        Sollten Sie Fragen haben, können Sie sich jederzeit an uns wenden:

        OneGov Cloud <br>
        Fabian Reinhard <br>
        Unter der Egg 5 <br>
        6004 Luzern <br>
        Tel. +41 41 511 22 50 <br>
        [fabian.reinhard@seantis.ch](fabian.reinhard@seantis.ch)

        ## Wie weiter

        Falls wir Sie überzeugen können melden Sie sich bei uns und wir
        können Ihre Testumgebung in eine produktive Umgebung überführen. Wir
        würden uns sehr freuen!
    """)

    pages = PageCollection(session)
    pages.add(
        parent=pages.by_path('aktuelles'),
        title=title,
        type='news',
        meta={
            'trait': 'page'
        },
        content={
            'lead': lead,
            'text': mistune.markdown(text, escape=False)
        }
    )


def add_builtin_forms(session, definitions=None):
    forms = FormCollection(session).definitions
    definitions = definitions or builtin_form_definitions()

    for name, title, definition in definitions:
        form = forms.by_name(name)

        if form:
            # update
            if form.title != title:
                form.title = title
            if form.definition != definition:
                form.definition = definition
        else:
            # add
            form = forms.add(
                name=name,
                title=title,
                definition=definition,
                type='custom'
            )

        assert form.form_class().has_required_email_field, (
            "Each form must have at least one required email field"
        )


def builtin_form_definitions(path=None):
    """ Yields the name, title and the form definition of all form definitions
    in the given or the default path.

    """
    path = path or module_path('onegov.town', 'forms/builtin')

    for filename in os.listdir(path):
        if filename.endswith('.form'):
            name = filename.replace('.form', '')
            title, definition = load_definition(os.path.join(path, filename))
            yield name, title, definition


def load_definition(path):
    """ Loads the title and the form definition from the given file. """

    with codecs.open(path, 'r', encoding='utf-8') as formfile:
        formlines = formfile.readlines()

        title = formlines[0].strip()
        definition = ''.join(formlines[3:])

        return title, definition


def add_resources(libres_context):
    resource = ResourceCollection(libres_context)
    resource.add(
        "SBB-Tageskarte",
        'Europe/Zurich',
        type='daypass',
        name='sbb-tageskarte'
    )


def add_events(session, town_name):
    start = as_datetime(datetime.today().date())
    while start.weekday() != 6:
        start = start + timedelta(days=1)

    events = EventCollection(session)
    event = events.add(
        title="150 Jahre {}".format(town_name),
        start=start + timedelta(hours=11, minutes=0),
        end=start + timedelta(hours=22, minutes=0),
        timezone="Europe/Zurich",
        tags=["Party"],
        location="Sportanlage",
        content={
            "description": "Lorem ipsum.",
            "organizer": "Gemeindeverwaltung"
        },
        meta={"submitter_email": "info@example.org"},
    )
    event.submit()
    event.publish()
    event = events.add(
        title="Gemeindeversammlung",
        start=start + timedelta(days=2, hours=20, minutes=0),
        end=start + timedelta(days=2, hours=22, minutes=30),
        timezone="Europe/Zurich",
        tags=["Politics"],
        location="Gemeindesaal",
        content={
            "description": "Lorem ipsum.",
            "organizer": "Gemeindeverwaltung"
        },
        meta={"submitter_email": "info@example.org"},
    )
    event.submit()
    event.publish()
    event = events.add(
        title="MuKi Turnen",
        start=start + timedelta(days=2, hours=10, minutes=0),
        end=start + timedelta(days=2, hours=11, minutes=0),
        recurrence=(
            "RRULE:FREQ=WEEKLY;WKST=MO;BYDAY=TU,TH;UNTIL={0}".format(
                (start + timedelta(days=31)).strftime('%Y%m%dT%H%M%SZ')
            )
        ),
        timezone="Europe/Zurich",
        tags=["Sports"],
        location="Turnhalle",
        content={
            "description": "Lorem ipsum.",
            "organizer": "Frauenverein"
        },
        meta={"submitter_email": "info@example.org"},
    )
    event.submit()
    event.publish()
    event = events.add(
        title="Grümpelturnier",
        start=start + timedelta(days=7, hours=10, minutes=0),
        end=start + timedelta(days=7, hours=18, minutes=0),
        timezone="Europe/Zurich",
        tags=["Sports"],
        location="Sportanlage",
        content={
            "description": "Lorem ipsum.",
            "organizer": "Sportverein"
        },
        meta={"submitter_email": "info@example.org"},
    )
    event.submit()
    event.publish()
